import numpy as np
from torch.utils.data.dataset import Dataset
import cmath
import copy
from utils import get_raw_data, get_base_model_output


class IAN_Dataset(Dataset):
    def __init__(self, dataset_name, split, base_model_name, rotate=True):
        super().__init__()
        self.dataset_name = dataset_name
        self.split = split
        self.base_model_name = base_model_name

        if dataset_name != "NBA":
            obs_len = 8
            pred_len = 12
        else:
            obs_len = 5
            pred_len = 10

        self.obs_len = obs_len
        self.pred_len = pred_len
        self.rotate = rotate

        self.feedback_list_obs = []
        self.feedback_list_gt = []
        self.feedback_base_pred_keys = []
        self.trj_order_keys = []
        self.base_model_outputs = None

        self.base_model_predictions, self.gt, self.base_prediction_match_table = get_base_model_output(base_model_name, dataset_name, split)
        raw_data, step_size = get_raw_data(dataset_name, split)

        for file_idx, trajectories in enumerate(raw_data):
            agent_ids = np.unique(trajectories[:, 1])
            for agent_id in agent_ids:
                agent_trj = trajectories[np.where(trajectories[:, 1] == agent_id)]
                if len(agent_trj) < obs_len + pred_len:
                    continue
                agent_past_trj = []
                agent_past_trj_valid_bit = []
                past_base_pred_key_list = []

                frame_begin, frame_end = int(agent_trj[0, 0]), int(agent_trj[-1, 0])
                continuous_cnt = 0
                for frame in range(frame_begin, frame_end + step_size, step_size):
                    agent_past_trj.append([])
                    agent_past_trj_valid_bit.append(np.zeros(obs_len + pred_len))
                    past_base_pred_key_list.append((file_idx, agent_id, frame))

                    frame_idx = np.where(agent_trj[:, 0] == frame)[0]
                    step_data = agent_trj[frame_idx]
                    if len(step_data) == 0:
                        continuous_cnt = 0
                        for i in range(len(agent_past_trj)):
                            if len(agent_past_trj[i]) < obs_len + pred_len:
                                agent_past_trj[i].append(np.zeros(2))
                    else:
                        continuous_cnt += 1
                        for i in range(len(agent_past_trj)):
                            if len(agent_past_trj[i]) < obs_len + pred_len:
                                agent_past_trj[i].append(step_data[0, 2:])
                                agent_past_trj_valid_bit[i][len(agent_past_trj[i]) - 1] = 1
                    if self.base_prediction_match_table.get(
                        (file_idx, agent_id, frame - step_size * (obs_len - 1)), None
                    ) is not None:  # Has enough steps for future, start collecting feedback
                        packed_feedback_obs = []
                        packed_feedback_gt = []
                        packed_base_pred_keys = []
                        for i in range(len(agent_past_trj)):
                            trj_piece = agent_past_trj[i]
                            valid_piece = agent_past_trj_valid_bit[i]
                            key = past_base_pred_key_list[i]

                            if np.sum(valid_piece) == obs_len + pred_len:
                                past = np.concatenate(
                                    [np.array(valid_piece[:obs_len])[..., None], np.array(trj_piece[:obs_len])],
                                    axis=1)
                                trj_gt = np.zeros((pred_len, 2))
                                actual_future_len = len(trj_piece[obs_len:])
                                assert actual_future_len == np.sum(valid_piece[obs_len:])
                                if actual_future_len > 0:
                                    trj_gt[:actual_future_len] = np.array(trj_piece[obs_len:])
                                future = np.concatenate([np.array(valid_piece[obs_len:])[..., None], trj_gt], axis=1)
                                if self.base_prediction_match_table.get(key, None) is None:
                                    continue
                                total_data = np.concatenate([past, future], axis=0)
                                assert len(total_data) == obs_len + pred_len
                                packed_feedback_obs.append(total_data[:obs_len])
                                packed_feedback_gt.append(total_data[obs_len:])
                                packed_base_pred_keys.append(key)

                        if len(packed_feedback_obs) > 0:
                            self.feedback_list_obs.append(packed_feedback_obs)
                            self.feedback_list_gt.append(packed_feedback_gt)
                            self.feedback_base_pred_keys.append(packed_base_pred_keys)
                            self.trj_order_keys.append((file_idx, agent_id, frame - (obs_len - 1) * step_size))

    def __len__(self):
        return len(self.feedback_list_obs)

    def __getitem__(self, item):
        trj_key = self.trj_order_keys[item]
        base_prediction_idx = int(self.base_prediction_match_table[trj_key])
        obs_trj = self.gt[base_prediction_idx][:self.obs_len]
        pred_gt = self.gt[base_prediction_idx][self.obs_len:]
        base_model_predictions = self.base_model_predictions[base_prediction_idx]

        ade_err = np.mean(np.linalg.norm(base_model_predictions - pred_gt[None], axis=-1), axis=-1)

        feedback_trj_past = copy.deepcopy(np.array(self.feedback_list_obs[item]))
        feedback_trj_future = copy.deepcopy(np.array(self.feedback_list_gt[item]))
        feedback_base_pred_keys = self.feedback_base_pred_keys[item]
        base_models_predictions_feedback = []

        for key in feedback_base_pred_keys:
            base_prediction_idx = int(self.base_prediction_match_table[key])
            base_model_predictions_fb = self.base_model_predictions[base_prediction_idx]
            base_models_predictions_feedback.append(base_model_predictions_fb)

        base_models_predictions_feedback = np.array(copy.deepcopy(base_models_predictions_feedback))

        if self.rotate:
            obs_offset = copy.deepcopy(obs_trj[-1:])
            obs_trj = obs_trj - obs_offset
            pred_gt = pred_gt - obs_offset
            base_model_predictions = base_model_predictions - obs_offset[None, ...]

            cmp = complex(-obs_trj[0, 0], -obs_trj[0, 1])
            _, theta = cmath.polar(cmp)
            matrix = np.array([[np.cos(-theta), -np.sin(-theta)],
                               [np.sin(-theta), np.cos(-theta)]])

            obs_trj = np.matmul(matrix, obs_trj.transpose()).transpose()
            pred_gt = np.matmul(matrix, pred_gt.transpose()).transpose()
            base_model_predictions = np.matmul(matrix, base_model_predictions.transpose((0, 2, 1))).transpose((0, 2, 1))

            for i in range(len(feedback_trj_past)):
                offset = copy.deepcopy(feedback_trj_past[i][-1:, 1:])
                feedback_trj_past[i][:, 1:] = feedback_trj_past[i][:, 1:] - offset
                feedback_trj_future[i][:, 1:] = feedback_trj_future[i][:, 1:] - offset * feedback_trj_future[i][:, :1]
                base_models_predictions_feedback[i] = base_models_predictions_feedback[i] - offset[None, ...]

                cmp = complex(-feedback_trj_past[i][0, 1], -feedback_trj_past[i][0, 2])
                _, theta = cmath.polar(cmp)
                matrix = np.array([[np.cos(-theta), -np.sin(-theta)],
                                   [np.sin(-theta), np.cos(-theta)]])

                feedback_trj_past[i][:, 1:] = np.matmul(matrix, feedback_trj_past[i][:, 1:].transpose()).transpose()
                feedback_trj_future[i][:, 1:] = np.matmul(matrix, feedback_trj_future[i][:, 1:].transpose()).transpose()
                base_models_predictions_feedback[i] = np.matmul(matrix, base_models_predictions_feedback[i].transpose(
                                                                        (0, 2, 1))).transpose((0, 2, 1))

        return obs_trj, pred_gt, base_model_predictions, \
               base_models_predictions_feedback, feedback_trj_past, feedback_trj_future, ade_err
