import copy
import numpy as np
import torch
import torch.nn as nn


class FeedbackGenerator(nn.Module):
    def __init__(self, lstm_feat_size=64, proposal_size=32, num_proposals=20):
        super(FeedbackGenerator, self).__init__()
        self.lstm_feat_size = lstm_feat_size

        self.obs_enc = nn.LSTM(2, lstm_feat_size, num_layers=2, batch_first=True)
        self.pred_enc = nn.LSTM(2, lstm_feat_size, num_layers=2, batch_first=True)
        self.gt_enc = nn.LSTM(2, lstm_feat_size, num_layers=2, batch_first=True)

        self.output_size = proposal_size * num_proposals * 2
        self.mlp = nn.Sequential(
            nn.Linear(lstm_feat_size * 2, lstm_feat_size * 4),
            nn.ReLU(),
            nn.Linear(lstm_feat_size * 4, lstm_feat_size * 8),
            nn.ReLU(),
            nn.Linear(lstm_feat_size * 8, self.output_size),
        )

    def initHidden(self, batch_size):
        return (torch.zeros(2, batch_size, self.trj_feat_size).cuda(),
                torch.zeros(2, batch_size, self.trj_feat_size).cuda())

    def forward(self, bs, feedback_trj, feedback_pred, feedback_sizes):
        batch_feedback_trj = torch.split(feedback_trj, feedback_sizes)
        size1, size2 = feedback_pred.shape[:2]
        feedback_pred_feat = []
        for i in range(size2):
            feedback_pred_part = feedback_pred[:, i]
            feedback_pred_hidden = self.initHidden(size1)
            feedback_pred_feat_part = self.pred_enc(feedback_pred_part, feedback_pred_hidden)[0][:, -1]
            feedback_pred_feat.append(feedback_pred_feat_part)
        feedback_pred_feat = torch.stack(feedback_pred_feat, dim=1)
        feedback_pred_feat = torch.split(feedback_pred_feat, feedback_sizes, dim=0)

        params = []
        for i in range(bs):
            feedback_trj_i = batch_feedback_trj[i]
            feedback_pred_feat_i = feedback_pred_feat[i]

            feedback_obs_hidden = self.initHidden(len(feedback_trj_i))
            feedback_obs_feat = self.obs_enc(feedback_trj_i[:, :self.obs_len], feedback_obs_hidden)[0][:, -1]
            feedback_gt_hidden = self.initHidden(len(feedback_trj_i))
            feedback_gt_feat = self.gt_enc(feedback_trj_i[:, self.obs_len:], feedback_gt_hidden)[0][:, -1]

            attn = torch.bmm(feedback_pred_feat_i, feedback_gt_feat[..., None]).squeeze(-1) / np.sqrt(feedback_pred_feat_i.shape[-1])
            attn_weight = torch.softmax(attn, dim=-1)
            agg_pred_feat = torch.max(feedback_pred_feat_i * attn_weight[..., None], dim=1)[0]
            feedback_feat = torch.cat([feedback_obs_feat, agg_pred_feat], dim=-1)

            param = self.mlp(feedback_feat)
            param = torch.max(param, dim=0)[0]
            params.append(param)

        params = torch.stack(params)

        return params[..., :self.output_size // 2], params[..., self.output_size // 2:]


class Adjuster(nn.Module):
    def __init__(self, obs_feat_size=64, proposal_size=32, num_proposals=20):
        super(Adjuster, self).__init__()
        self.obs_feat_size = obs_feat_size
        self.proposal_size = proposal_size
        self.num_proposals = num_proposals

        input_size = self.obs_feat_size
        self.mlp_1 = nn.Sequential(
            nn.Linear(input_size, input_size * 2),
            nn.ReLU(),
            nn.Linear(input_size * 2, input_size * 4),
            nn.ReLU(),
            nn.Linear(input_size * 4, self.num_proposals * self.proposal_size, bias=False)
        )
        self.mlp_2 = nn.Linear(self.num_proposals * self.proposal_size,
                               self.num_proposals * self.proposal_size, bias=False)

    def forward(self, obs_feat, param_1, param_2):
        bs = len(obs_feat)
        data = self.mlp_1(obs_feat)
        data += param_1
        data = torch.relu(data)
        data = self.mlp_2(data)
        data += param_2
        res = data.reshape(bs, self.num_proposals, self.proposal_size)
        return res


class ConfidenceEstimator(nn.Module):
    def __init__(self, proposal_size=32, trj_feat_size=64):
        super(ConfidenceEstimator, self).__init__()
        self.proposal_size = proposal_size
        self.trj_feat_size = trj_feat_size

        self.trj_encoder = nn.LSTM(2, trj_feat_size, num_layers=2, batch_first=True)
        input_size = proposal_size + trj_feat_size
        self.mlp = nn.Sequential(
            nn.Linear(input_size, input_size),
            nn.ReLU(),
            nn.Linear(input_size, input_size),
            nn.ReLU(),
            nn.Linear(input_size, 1),
        )

    def initHidden(self, batch_size):
        return (torch.zeros(2, batch_size, self.trj_feat_size).cuda(),
                torch.zeros(2, batch_size, self.trj_feat_size).cuda())

    def forward(self, proposal, trj_pool):
        num_proposal = len(proposal)
        num_trj = len(trj_pool)
        trj_hidden = self.initHidden(num_trj)
        trj_pool_feat = self.trj_encoder(trj_pool, trj_hidden)[0][:, -1]

        proposal_repeat = proposal[:, None].repeat(1, num_trj, 1)
        trj_pool_repeat = trj_pool_feat[None].repeat(num_proposal, 1, 1)
        confidence = self.mlp(torch.cat([proposal_repeat, trj_pool_repeat], dim=-1)).squeeze(-1)
        return confidence


class IAN(nn.Module):
    def __init__(self, lstm_feat_size=64, obs_len=8, pred_len=12, topk=20, eta=200):
        super(IAN, self).__init__()
        self.obs_len = obs_len
        self.pred_len = pred_len
        self.lstm_feat_size = lstm_feat_size
        self.topk = topk
        self.eta = eta
        self.prediction_model = None

        # Current Encoder
        self.current_obs_enc = nn.LSTM(2, lstm_feat_size, num_layers=2, batch_first=True)
        self.current_pred_enc = nn.LSTM(2, lstm_feat_size, num_layers=2, batch_first=True)

        self.feedback_gen = FeedbackGenerator()
        self.adjuster = Adjuster(obs_feat_size=lstm_feat_size * 2)
        self.confidence_network = ConfidenceEstimator()

    def initHidden(self, batch_size):
        return (torch.zeros(2, batch_size, self.lstm_feat_size).cuda(),
                torch.zeros(2, batch_size, self.lstm_feat_size).cuda())

    def adjust_train(self, obs_trj, base_model_output, base_pred_ade, feedback_trj, feedback_pred, feedback_sizes):
        bs = len(obs_trj)
        current_obs_hidden = self.initHidden(bs)
        current_obs_feat = self.current_obs_enc(obs_trj, current_obs_hidden)[0][:, -1]

        current_pred = base_model_output[:, :self.topk].reshape(-1, self.pred_len, 2)
        current_pred_hidden = self.initHidden(bs * self.topk)
        current_pred_feat = self.current_pred_enc(current_pred, current_pred_hidden)[0][:, -1]
        current_pred_feat = current_pred_feat.reshape(bs, self.topk, -1)
        current_pred_feat = torch.max(current_pred_feat, dim=1)[0]

        param_1, param_2 = self.feedback_gen(bs, feedback_trj, feedback_pred, feedback_sizes)
        proposals = self.adjuster(torch.cat([current_obs_feat, current_pred_feat], dim=-1), param_1, param_2)
        expectation_loss = 0
        for i in range(bs):
            confidence = self.confidence_network(proposals[i], base_model_output[i])
            confidence_prob = torch.softmax(confidence, dim=-1)
            min_expectation = (confidence_prob * base_pred_ade[i][None]).sum(dim=-1).min()
            expectation_loss += min_expectation

        loss = expectation_loss / bs
        return loss

    @torch.no_grad()
    def adjust(self, obs_trj, base_model_output, feedback_trj, feedback_pred, feedback_sizes):
        bs = len(obs_trj)
        current_obs_hidden = self.initHidden(bs)
        current_obs_feat = self.current_obs_enc(obs_trj, current_obs_hidden)[0][:, -1]

        current_pred = base_model_output[:, :self.topk].reshape(-1, self.pred_len, 2)
        current_pred_hidden = self.initHidden(bs * self.topk)
        current_pred_feat = self.current_pred_enc(current_pred, current_pred_hidden)[0][:, -1]
        current_pred_feat = current_pred_feat.reshape(bs, self.topk, -1)
        current_pred_feat = torch.max(current_pred_feat, dim=1)[0]

        param_1, param_2 = self.feedback_gen(bs, feedback_trj, feedback_pred, feedback_sizes)
        proposals = self.adjuster(torch.cat([current_obs_feat, current_pred_feat], dim=-1), param_1, param_2)

        # Query & Filter
        batch_filtered_trj = []
        batch_candidate_predictions = self.query_predictions_from_base_model()  # bs * \eta * pred_len * 2

        for i in range(bs):
            filtered_candidates = []
            confidence = self.confidence_network(proposals[i], batch_candidate_predictions[i]).cpu()  # k * \eta
            max_idx = torch.argmax(confidence).item()
            idx_1 = max_idx // self.eta
            idx_2 = max_idx % self.eta

            while confidence[idx_1, idx_2] > -torch.inf:
                filtered_candidates.append(batch_candidate_predictions[i][idx_2])
                confidence[idx_1, :] = -torch.inf
                confidence[:, idx_2] = -torch.inf
                max_idx = torch.argmax(confidence).item()
                idx_1 = max_idx // self.eta
                idx_2 = max_idx % self.eta

            filtered_candidates = torch.stack(filtered_candidates)
            batch_filtered_trj.append(filtered_candidates)

        return batch_filtered_trj

    def set_base_model(self, model_name):
        pass

    def query_predictions_from_base_model(self):
        pass
